#------------------------------------------------------------------
# $Copyright: Copyright (c) 2009 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/PowerCenterSvcMgr/PowerCenterSvcMgr.pm,v 1.1.2.7 2009/05/25 08:09:19 aatre Exp $
#
# $Author: aatre $
# $Date: 2009/05/25 08:09:19 $
# $Revision: 1.1.2.7 $
#
# Provides:                            PowerCenterSvcMgr
# Author:                              Anshuman Atre
# crAgent Build:                       1.87 2008/10/13 06:58:52
# crAgent Run Date:                    13 Oct, 2008
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package PowerCenterSvcMgr;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: PowerCenterSvcMgr.pm,v 1.1.2.7 2009/05/25 08:09:19 aatre Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define PowerCenterSvcMgr subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String InfaUser,
                              IN String EnvFile,
                              IN String HostName,
                              IN Integer Port,
                              IN String InfaHome,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localInfaUser                 The dedicated user for the Service Manager
 $localEnvFile                  Absolute path to environment file to source
 $localHostName                 Hostname/IPv4 address of the virtual host
 $localPort                     The port on which the Service Manager listens
 $localInfaHome                 Absolute path to installation directory of instance
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localInfaUser,
       $localEnvFile,
       $localHostName,
       $localPort,
       $localInfaHome,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  InfaUser                 [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  Port                     [%s]
  InfaHome                 [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,   $sEntryPointName,   $localInfaUser,    $localEnvFile,
  $localHostName,            $localPort,            $localInfaHome,
  $localSecondLevelMonitor, $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName   InfaUser   EnvFile   HostName   Port    InfaHome
       SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile HostName InfaHome InfaUser Port );
  my @lsOfflineArgs = qw ( EnvFile HostName InfaHome InfaUser Port );
  my @lsMonitorArgs = qw ( InfaHome InfaUser );
  my @lsCleanArgs   = qw ( HostName InfaHome InfaUser Port );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #-----------------------------------------------
  # Validation for online offline entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20003, "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------
    # Validate parameter InfaHome..
    #------------------------------
    if ( ! -d $localInfaHome ) {
      &logError ( 20004, "<%s> InfaHome [%s] directory does not exist",
      $sMyName, $localInfaHome );
      $iReturnCode = 0; # Failure..
    }
  }

  #----------------------------------
  # Validation for all entry points..
  #----------------------------------
  #----------------------------
  # Validate parameter InfaUser..
  #----------------------------
  if ( ! defined getpwnam $localInfaUser ) {
     &logError ( 20005, "<%s> InfaUser [%s] is not a valid UNIX login",
     $sMyName, $localInfaUser );
     $iReturnCode = 0; # Failure..
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20006, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  #-----------------------------------------------
  # Validation for online monitor entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|monitor)$/o ) {
    #--------------------------
    # Validate parameter Port..
    #--------------------------
    if ( $localSecondLevelMonitor && ! $localPort ) {
      &logError ( 20007,
      "<%s> Port [%s] has to be greater than 0 for SecondLevelMonitor [%s]",
      $sMyName, $localPort, $localSecondLevelMonitor );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------
    # Validate parameter HostName..
    #------------------------------
    if ( ! inet_aton $localHostName ) {
      &logError ( 20008,
      "<%s> HostName [%s] is an invalid Hostname/IPv4 address",
      $sMyName, $localHostName );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20009,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String EnvFile,
                         IN String HostName,
                         IN String InfaHome,
                         IN String InfaUser,
                         IN Integer Port,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaHome                 Absolute path to installation directory of instance
 $sInfaUser                 The dedicated user for the Service Manager
 $iPort                     The port on which the Service Manager listens
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sHostName,
       $sInfaHome,
       $sInfaUser,
       $iPort,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20010, "Subroutine <%s> called with:
  EnvFile                  [%s]
  HostName                 [%s]
  InfaHome                 [%s]
  InfaUser                 [%s]
  Port                     [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $sEnvFile, $sHostName, $sInfaHome, $sInfaUser,  $iPort,
  $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting PowerCenterSvcMgr processes..
  #-----------------------------------------
  &logTrace ( 20011,
  "<%s> Killing PowerCenterSvcMgr processes for this instance", $sMyName );

  if ( PowerCenterSvcMgr::KillPowerCenterSvcMgrProcesses ( $sHostName,
                                                           $iPort,
                                                           $sInfaHome,
                                                           $sInfaUser ) ) {
    &logTrace ( 20012,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20013,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + PowerCenterSvcMgr::GetEngineTimeout ( 1 );
  
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  $sCommand = "$sInfaHome/server/tomcat/bin/infaservice.sh";
  $sCommandArgs = "startup";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20014,
  "<%s> Starting PowerCenterSvcMgr instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20015,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    PowerCenterSvcMgr::ExecLocalCmd ( $sInfaUser, 
                                      $sEnvFile, 
                                      $sCommand, 
                                      $sCommandArgs, 
                                      \$iExitCode, 
                                      $iTimeout, 
                                      \@lsCmdout ) ) {
    &logTrace ( 20016,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20017,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         PowerCenterSvcMgr::GetProcState ( \@liPids,
                                           $sHostName,
                                           $iPort,
                                           $sInfaHome,
                                           $sInfaUser,
                                           1 ) && (         # Running FFLM..
         $iSecondLevelMonitor  ?                            # If enabled..
         VCSDef_MonitorExit::ONLINE ==                      # Online after..
         PowerCenterSvcMgr::SecondLevelMonitor ( $sEnvFile, # Running SLM..
                                                 $sHostName,
                                                 $sInfaUser,
                                                 $iPort,
                                                 $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20018,
      "<%s> Confirmed start-up of this instance of PowerCenterSvcMgr server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20019,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String EnvFile,
                        IN String HostName,
                        IN String InfaHome,
                        IN String InfaUser,
                        IN Integer Port );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaHome                 Absolute path to installation directory of instance
 $sInfaUser                 The dedicated user for the Service Manager
 $iPort                     The port on which the Service Manager listens
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sHostName,
       $sInfaHome,
       $sInfaUser,
       $iPort ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20020, "Subroutine <%s> called with:
  EnvFile                  [%s]
  HostName                 [%s]
  InfaHome                 [%s]
  InfaUser                 [%s]
  Port                     [%s]",

  $sMyName, $sEnvFile, $sHostName,
  $sInfaHome, $sInfaUser, $iPort );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + PowerCenterSvcMgr::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  $sCommand = "$sInfaHome/server/tomcat/bin/infaservice.sh";
  $sCommandArgs = "shutdown";
  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20021,
  "<%s> Stopping PowerCenterSvcMgr with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20022,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          PowerCenterSvcMgr::ExecLocalCmd ( $sInfaUser, 
                                            $sEnvFile, 
                                            $sCommand, 
                                            $sCommandArgs, 
                                            \$iExitCode, 
                                            $iTimeout, 
                                            \@lsCmdout ) ) {
    &logTrace ( 20023,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20024,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      PowerCenterSvcMgr::GetProcState ( \@liPids,
                                        $sHostName,
                                        $iPort,
                                        $sInfaHome,
                                        $sInfaUser,
                                        0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20025,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20026,
        "<%s> Confirmed shutdown of PowerCenterSvcMgr instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #---------------------------------------------------
  # Whack any persisting PowerCenterSvcMgr processes..
  #---------------------------------------------------
  &logTrace ( 20027,
  "<%s> Killing PowerCenterSvcMgr processes for this instance", $sMyName );

  if ( PowerCenterSvcMgr::KillPowerCenterSvcMgrProcesses ( $sHostName,
                                                           $iPort,
                                                           $sInfaHome,
                                                           $sInfaUser ) ) {
    &logTrace ( 20028,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20029,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String EnvFile,
                        IN String HostName,
                        IN String InfaHome,
                        IN String InfaUser,
                        IN Integer Port );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaHome                 Absolute path to installation directory of instance
 $sInfaUser                 The dedicated user for the Service Manager
 $iPort                     The port on which the Service Manager listens
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sEnvFile,
       $sHostName,
       $sInfaHome,
       $sInfaUser,
       $iPort ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20030, "Subroutine <%s> called with:
  CleanReason              [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  InfaHome                 [%s]
  InfaUser                 [%s]
  Port                     [%s]",

  $sMyName,  $sCleanReason,  $sEnvFile,   $sHostName,   $sInfaHome,
  $sInfaUser, $iPort );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + PowerCenterSvcMgr::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sInfaHome/server/tomcat/bin/infaservice.sh";
  $sCommandArgs = "shutdown";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20031,
    "<%s> Attempting to stop PowerCenterSvcMgr with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #-------------------
    # Request shutdown..
    #-------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20032,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( PowerCenterSvcMgr::ExecLocalCmd ( $sInfaUser, 
                                              $sEnvFile, 
                                              $sCommand, 
                                              $sCommandArgs, 
                                              \$iExitCode, 
                                              $iTimeout, 
                                              \@lsCmdout ) ) {
      &logTrace ( 20033, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #---------------------------------------------------
  # Whack any persisting PowerCenterSvcMgr processes..
  #---------------------------------------------------
  &logTrace ( 20034,
  "<%s> Killing PowerCenterSvcMgr processes for this instance", $sMyName );

  if ( PowerCenterSvcMgr::KillPowerCenterSvcMgrProcesses ( $sHostName,
                                                           $iPort,
                                                           $sInfaHome,
                                                           $sInfaUser ) ) {
    &logTrace ( 20035,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20036,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String HostName,
                                   IN Integer Port,
                                   IN String InfaHome,
                                   IN String InfaUser );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sHostName                 Hostname/IPv4 address of the virtual host
 $iPort                     The port on which the Service Manager listens
 $sInfaHome                 Absolute path to installation directory of instance
 $sInfaUser                 The dedicated user for the Service Manager
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sHostName,
       $iPort,
       $sInfaHome,
       $sInfaUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20037, "Subroutine <%s> called:
  HostName                 [%s]
  Port                     [%s]
  InfaHome                 [%s]
  InfaUser                 [%s]",

  $sMyName, $sHostName, $iPort, $sInfaHome, $sInfaUser );
  
  #------------------------------------------
  # Check if virtual hostname is up locally..
  #------------------------------------------
  unless ( VCSagentFW::RunGetProcState ( $sHostName ) ) {
    &logTrace ( 20038,
    "<%s> Virtual Host [%s] not up locally", $sMyName, $sHostName );
    return VCSDef_MonitorExit::OFFLINE;
  }

  my @liDummyPids = ();
  my $iReturnCode = PowerCenterSvcMgr::GetProcState ( \@liDummyPids,
                                                      $sHostName,
                                                      $iPort,
                                                      $sInfaHome,
                                                      $sInfaUser,
                                                      1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20039,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20040,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20041,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String HostName,
                               IN Integer Port,
                               IN String InfaHome,
                               IN String InfaUser,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sHostName                 Hostname/IPv4 address of the virtual host
 $iPort                     The port on which the Service Manager listens
 $sInfaHome                 Absolute path to installation directory of instance
 $sInfaUser                 The dedicated user for the Service Manager
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sHostName,
       $iPort,
       $sInfaHome,
       $sInfaUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20042, "Subroutine <%s> called with:
  OnlinePids               [%s]
  HostName                 [%s]
  Port                     [%s]
  InfaHome                 [%s]
  InfaUser                 [%s]
  UseFFLM                  [%s]",

  $sMyName,  $rliOnlinePids,  $sHostName,  $iPort, $sInfaHome,  $sInfaUser,
  $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  
  #--------------------------
  # Other processes to kill..
  #--------------------------
  my @lsOthers = ( qr!$sInfaHome\b!,
                   qr!http://$sHostName:$iPort! );

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sInfaUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                      # And called to kill-by-UID..
    &logTrace ( 20043, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sInfaUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( PowerCenterSvcMgr::GetSrvrDetails ( $sHostName, 
                                             $sInfaHome, 
                                             \%hsSrvrDetails ) ) {
      &logTrace ( 20044,
      "<%s> Got details for monitoring PowerCenterSvcMgr instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20045,
      "<%s> Failed to get details for monitoring PowerCenterSvcMgr instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {                     # Use copy..
        @lsUpdates = ();                                   # Reset..
        @lsValues = @{ $hsCopy { $sKey } };                # Get values..
        $sTemp .= "\nPowerCenterSvcMgr process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = PowerCenterSvcMgr::FFLMStatus ( $sPidFile,
                                                    $sProc,
                                                    $iUid,
                                                    \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20046,
      "<%s> PowerCenterSvcMgr instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20047,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sInfaUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20048,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20049,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20050,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20051, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                  unshift ( @$rliOnlinePids, $iPid );         # Store..
    
                  &logTrace ( 20052,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  $hiProcPids { $sKey } = $iPid;        # Store..
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20053,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( ! grep { /^$iPid$/ } @$rliOnlinePids ) &&         # Not caught earlier..
           map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..
        push ( @$rliOnlinePids, $iPid );                       # Store..

        &logTrace ( 20054,
        "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20055,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( PowerCenterSvcMgr::CreatePidFile ( $sPidFile,
                                                $hiProcPids { $sKey } ) ) {
          &logTrace ( 20056,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20057,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20058,
    "<%s> Found [%s] PIDs [%s] for this instance of PowerCenterSvcMgr.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20059,
    "<%s> PowerCenterSvcMgr process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String HostName,
                          IN String InfaUser,
                          IN String MonitorProgram,
                          IN Integer Port,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaUser                 The dedicated user for the Service Manager
 $sMonitorProgram           Full path to external monitor program
 $iPort                     The port on which the Service Manager listens
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sHostName,
       $sInfaUser,
       $sMonitorProgram,
       $iPort,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20060, "Subroutine <%s> called with:
  EnvFile                  [%s]
  HostName                 [%s]
  InfaUser                 [%s]
  MonitorProgram           [%s]
  Port                     [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $sEnvFile,  $sHostName,  $sInfaUser,  $sMonitorProgram,
  $iPort, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         PowerCenterSvcMgr::DelayedAttributeCheck ( $sEnvFile,
                                                    $sHostName,
                                                    $sInfaUser,
                                                    $iPort ) ) ) {
      &logTrace ( 20061,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20062,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20063,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + PowerCenterSvcMgr::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         PowerCenterSvcMgr::SecondLevelMonitor ( $sEnvFile,
                                                 $sHostName,
                                                 $sInfaUser,
                                                 $iPort,
                                                 $iTillWhen ) ) ) {

      &logTrace ( 20064,
      "<%s> PowerCenterSvcMgr instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20065,
      "<%s> PowerCenterSvcMgr instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + PowerCenterSvcMgr::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         PowerCenterSvcMgr::ExternalMonitor ( $sInfaUser,
                                              $sEnvFile,
                                              $sExtCmd,
                                              $sExtCmdArgs,
                                              $iTillWhen ) ) ) {
      &logTrace ( 20066,
      "<%s> PowerCenterSvcMgr instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20067,
      "<%s> PowerCenterSvcMgr instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String HostName,
                                        IN String InfaUser,
                                        IN Integer Port );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaUser                 The dedicated user for the Service Manager
 $iPort                     The port on which the Service Manager listens
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sHostName,
       $sInfaUser,
       $iPort ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20068, "Subroutine <%s> called with:
  EnvFile                  [%s]
  HostName                 [%s]
  InfaUser                 [%s]
  Port                     [%s]",

  $sMyName, $sEnvFile, $sHostName, $sInfaUser, $iPort );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EnvFile HostName InfaUser Port );
  my @lsReqdArgs = qw ( EnvFile HostName InfaUser Port );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20069,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20070, "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-------------------------------
  # Validate parameter InfaUser..
  #-------------------------------
  if ( ! defined getpwnam $sInfaUser ) {
    &logError ( 20071, "<%s> InfaUser [%s] is not a valid UNIX login",
    $sMyName, $sInfaUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #---------------------------
  # Validate parameter Port..
  #---------------------------
  if ( ! $iPort ) {
    &logError ( 20072, "<%s> Port [%s] has to be greater than 0",
    $sMyName, $iPort );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-------------------------------
  # Validate parameter HostName..
  #-------------------------------
  if ( ! inet_aton $sHostName ) {
    &logError ( 20073, "<%s> HostName [%s] is an invalid Hostname/IPv4 address",
    $sMyName, $sHostName );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String HostName,
                                     IN String InfaUser,
                                     IN Integer Port,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaUser                 The dedicated user for the Service Manager
 $iPort                     The port on which the Service Manager listens
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sHostName,
       $sInfaUser,
       $iPort,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20074, "Subroutine <%s> called with:
  EnvFile                  [%s]
  HostName                 [%s]
  InfaUser                 [%s]
  Port                     [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile, $sHostName, $sInfaUser, $iPort, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #------------------------------------
  # Check if port can be connected to..
  #------------------------------------
  if ( Comms::CheckPort ( $sHostName,
                          $iPort ) ) {
    &logTrace ( 20075,
    "<%s> Server listening on Host [%s] Port [%s]: Socket method",
    $sMyName, $sHostName, $iPort );
  }
  else {
    &logError ( 20076,
    "<%s> Server NOT listening on Host [%s] Port [%s]: Socket method",
    $sMyName, $sHostName, $iPort );

    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String InfaUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sInfaUser                 The dedicated user for the Service Manager
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInfaUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20077, "Subroutine <%s> called with:
  InfaUser                 [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sInfaUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20078,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20079,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20080,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20081,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20082,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  PowerCenterSvcMgr::ExecLocalCmd ( $sInfaUser, 
                                    $sEnvFile, 
                                    $sCmd, 
                                    $sCmdArgs, 
                                    \$iExitCode, 
                                    $iTimeout, 
                                    \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20083,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20084,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20085, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20086, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20087,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20088,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20089,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20090, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20091, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20092, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20093, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20094,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20095,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20096,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( PowerCenterSvcMgr::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20097,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20098,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20099,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20100,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20101, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillPowerCenterSvcMgrProcesses ( IN String HostName,
                                                  IN Integer Port,
                                                  IN String InfaHome,
                                                  IN String InfaUser );

 Description  : Kills processes for this instance of PowerCenterSvcMgr
 
 Parameters :

 $sHostName               Hostname/IPv4 address of the virtual host
 $iPort                     The port on which the Service Manager listens
 $sInfaHome               Absolute path to installation directory of instance
 $sInfaUser               The dedicated user for the Service Manager
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillPowerCenterSvcMgrProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sHostName,
       $iPort,
       $sInfaHome,
       $sInfaUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20102, "Subroutine <%s> called with:
  HostName                 [%s]
  Port                     [%s]
  InfaHome                 [%s]
  InfaUser                 [%s]",

  $sMyName, $sHostName, $iPort, $sInfaHome, $sInfaUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'PowerCenterSvcMgr::Start' ) { # During online..
    @liGoners = @liRunningProcs;                         # Get PIDs to kill..
  }
  else {
    PowerCenterSvcMgr::GetProcState ( \@liGoners,
                                      $sHostName,
                                      $iPort,
                                      $sInfaHome,
                                      $sInfaUser,
                                      0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20103,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20104,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20105,
    "<%s> Could not locate any process this instance of PowerCenterSvcMgr",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20106, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20107,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20108,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20109,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20110,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20111,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20112,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20113,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20114,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20115, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20116,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  goto $sSwitch; {
    solaris:

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;

    hpux:

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20117, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20118,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20119,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String HostName,
                                  IN String InfaHome,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sHostName                 Hostname/IPv4 address of the virtual host
 $sInfaHome                 Absolute path to installation directory of instance
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sHostName,
       $sInfaHome,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20120, "Subroutine <%s> called with:
  HostName                 [%s]
  InfaHome                 [%s]
  SrvrDetails              [%s]",

  $sMyName, $sHostName, $sInfaHome, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20121, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20122, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  $sProc = qr!$sInfaHome/java/bin/java.*java\.awt\.headless=true\b!;
  $sKey = 'headless';
  $sPidFile = "$sPIDDir/pidfile.$sKey";
  @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20123, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  $sCommand = $sCmdHares; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceName $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                    =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20124,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
  }
  elsif ( ! $iEngineTimeout ) { # Undefined value..
    &logTrace ( 20125, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
  }
  else { # Success..
    &logTrace ( 20126, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  unless ( $iEngineTimeout ) {
    $sCommand = $sCmdHatype; # Get global..
    ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
    $sResourceType = "PowerCenterSvcMgr" if $sResourceType !~ /\w/; # Ensure..
    $sCommandArgs = "-value $sResourceType $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20127,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    elsif ( ! $iEngineTimeout ) { # Empty value..
      &logTrace ( 20128, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
      $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
    }
    else { # Success..
      &logTrace ( 20129, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20130,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

1; # So PERL knows its included..
